--==============================================================================
-- CERN (BE-CO-HT)
-- Converter board common gateware top-level file
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2014-07-24
--
-- version: 1.0
--
-- description:
--    This module is to be instantiated in all pulse converter board designs,
--    for example the CONV-TTL-BLO or CONV-TTL-RS485. It contains a set of
--    modules common to all these boards, that are configurable via generics to
--    accommodate for the application of each board.
--
-- dependencies:
--    general-cores repository [1]
--
-- references:
--   [1] Platform-independent core collection webpage on OHWR,
--       http://www.ohwr.org/projects/general-cores/repository
--   [2] ELMA, Access to board data using SNMP and I2C
--       http://www.ohwr.org/documents/227
--
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2014-07-24   Theodor Stana     File created
--==============================================================================
-- TODO: -
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.genram_pkg.all;

entity conv_common_gw is
  generic
  (
    -- Number of repeater channels
    g_nr_chans            : integer := 6;

    -- Gateware version
    g_gwvers              : std_logic_vector(7 downto 0);

    -- Generate pulse repetition logic with fixed output pulse width
    g_pgen_fixed_width    : boolean;
    -- Pulse width at pulse generator output (valid with fixed output pulse width)
    g_pgen_pwidth         : natural range 20 to 40 := 24;
    -- Duty cycle divider ratio for pulse generator
    -- output pulse will be limited to 1/g_pgen_duty_cycle_div
    g_pgen_duty_cycle_div : natural := 5;
    -- Pulse generator glitch filter length in number of clk_20_i cycles
    g_pgen_gf_len         : integer := 4;

    -- Bicolor LED controller signals
    g_bicolor_led_columns : integer := 6;
    g_bicolor_led_lines   : integer := 2
  );
  port
  (
    -- Clocks
    clk_20_i                : in  std_logic;
    clk_125_p_i             : in  std_logic;
    clk_125_n_i             : in  std_logic;

    -- Reset output signal, synchronous to 20 MHz clock
    rst_n_o                 : out std_logic;

    -- Glitch filter active-low enable signal
    gf_en_n_i               : in  std_logic;

    -- Channel enable
    global_ch_oen_o         : out std_logic;
    pulse_front_oen_o       : out std_logic;
    pulse_rear_oen_o        : out std_logic;
    inv_oen_o               : out std_logic;

    -- Pulse inputs
    pulse_i                 : in  std_logic_vector(g_nr_chans-1 downto 0);
    pulse_o                 : out std_logic_vector(g_nr_chans-1 downto 0);

    -- Channel leds
    led_pulse_o             : out std_logic_vector(g_nr_chans-1 downto 0);

    -- Bicolor LED signals
    bicolor_led_state_i     : in  std_logic_vector(2*g_bicolor_led_columns*g_bicolor_led_lines-1 downto 0);
    bicolor_led_col_o       : out std_logic_vector(g_bicolor_led_columns-1 downto 0);
    bicolor_led_line_o      : out std_logic_vector(g_bicolor_led_lines-1 downto 0);
    bicolor_led_line_oen_o  : out std_logic_vector(g_bicolor_led_lines-1 downto 0)
  );
end entity conv_common_gw;


architecture arch of conv_common_gw is

  --============================================================================
  -- Type declarations
  --============================================================================
  type t_pulse_led_cnt is array(0 to g_nr_chans-1) of unsigned(18 downto 0);
  type t_pulse_cnt is array(0 to g_nr_chans-1) of unsigned(31 downto 0);
  type t_ch_pcr is array(0 to g_nr_chans-1) of std_logic_vector(31 downto 0);

  --============================================================================
  -- Constant declarations
  --============================================================================
  -- Number of Wishbone masters and slaves, for wb_crossbar
  constant c_nr_masters      : natural :=  1;
  constant c_nr_slaves       : natural :=  3;

  -- slave order definitions
  constant c_slv_conv_regs    : natural := 0;
  constant c_slv_multiboot    : natural := 1;
  constant c_slv_onewire_mst  : natural := 2;

  -- base address definitions
  constant c_addr_conv_regs   : t_wishbone_address := x"00000000";
  constant c_addr_multiboot   : t_wishbone_address := x"00000040";
  constant c_addr_onewire_mst : t_wishbone_address := x"00000080";

  -- address mask definitions
  constant c_mask_conv_regs   : t_wishbone_address := x"00000fc0";
  constant c_mask_multiboot   : t_wishbone_address := x"00000fc0";
  constant c_mask_onewire_mst : t_wishbone_address := x"00000fe0";

  -- addresses constant for Wishbone crossbar
  constant c_addresses : t_wishbone_address_array(c_nr_slaves-1 downto 0)
                  := (
                       c_slv_conv_regs   => c_addr_conv_regs,
                       c_slv_multiboot   => c_addr_multiboot,
                       c_slv_onewire_mst => c_addr_onewire_mst
                     );

  -- masks constant for Wishbone crossbar
  constant c_masks : t_wishbone_address_array(c_nr_slaves-1 downto 0)
                  := (
                       c_slv_conv_regs   => c_mask_conv_regs,
                       c_slv_multiboot   => c_mask_multiboot,
                       c_slv_onewire_mst => c_mask_onewire_mst
                     );


  --============================================================================
  -- Component declarations
  --============================================================================
  ------------------------------------------------------------------------------
  -- Fixed-width reset generator
  ------------------------------------------------------------------------------
  component conv_reset_gen is
    generic
    (
      -- Reset time in number of clk_i cycles
      g_reset_time : positive := 2_000_000
    );
    port
    (
      clk_i   : in  std_logic;
      rst_i   : in  std_logic;
      rst_n_o : out std_logic
    );
  end component conv_reset_gen;

  component conv_pulse_gen is
    generic
    (
      -- This generic enables elaboration of the fixed pulse width logic
      g_with_fixed_pwidth : boolean;

      -- Pulse width, in number of clk_i cycles
      -- Default pulse width (20 MHz clock): 1.2 us
      -- Minimum allowable pulse width (20 MHz clock): 1 us
      -- Maximum allowable pulse width (20 MHz clock): 2 us
      g_pwidth : natural range 20 to 40 := 24;

      -- Duty cycle divider: D = 1/g_duty_cycle_div
      g_duty_cycle_div : natural := 5
    );
    port
    (
      -- Clock and active-low reset inputs
      clk_i         : in  std_logic;
      rst_n_i       : in  std_logic;

      -- Glitch filter enable input
      -- '1' - Glitch filter disabled (glitch-sensitive, no output jitter)
      -- '0' - Glitch filter enabled (glitch-insensitive, with output jitter)
      gf_en_n_i     : in  std_logic;

      -- Enable input, pulse generation is enabled when '1'
      en_i          : in  std_logic;

      -- Trigger input, has to be '1' to assure pulse output with delay no greater
      -- than internal gate delays.
      trig_a_i      : in  std_logic;

      -- Pulse error output, pulses high for one clock cycle when a pulse arrives
      -- within a pulse period
      pulse_err_p_o : out std_logic;

      -- Pulse output, active-high
      -- latency:
      --    glitch filter disabled: none
      --    glitch filter enabled: glitch filter length + 5 clk_i cycles
      pulse_o       : out std_logic
    );
  end component conv_pulse_gen;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Per-domain clock and reset signals
  signal clk_125            : std_logic;
  signal rst_125_n          : std_logic;
  signal rst_20_n           : std_logic;

  -- Pulse logic signals
  signal trig_a             : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_synced        : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_degl          : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_chan          : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_pgen          : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp         : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp_d0      : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp_redge_p : std_logic_vector(g_nr_chans-1 downto 0);

  signal led_pulse          : std_logic_vector(g_nr_chans-1 downto 0);
  signal led_pulse_cnt      : t_pulse_led_cnt;

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Internal and external reset generation
  --============================================================================
  -- Configure reset generator for 100ms power-on reset
  cmp_reset_gen : conv_reset_gen
    generic map
    (
      -- Reset time: 50ns * 2 * (10**6) = 100 ms
      g_reset_time => 2*(10**6)
    )
    port map
    (
      clk_i   => clk_20_i,
      rst_i   => '0',
      rst_n_o => rst_20_n
    );

  -- Output reset signal is synchronous to the 20 MHz clock
  rst_n_o <= rst_20_n;

  --============================================================================
  -- Output enable logic
  --============================================================================
  global_ch_oen_o   <= '1';
  pulse_front_oen_o <= '1';
  inv_oen_o         <= '1';
  pulse_rear_oen_o  <= '1';

  --============================================================================
  -- Pulse repetition logic
  --============================================================================
  trig_a <= pulse_i;

  gen_pulse_chan_logic : for i in 0 to g_nr_chans-1 generate

    -- Synchronize the asynchronous trigger input into the 20 MHz clock
    -- domain before passing it to the glitch filter
    --
    -- Reset value is '1' to avoid pulses being counted by pulse counter on
    -- startup, when the board is in TTL-BAR repetition mode.
    cmp_trig_sync : gc_sync_ffs
      generic map
      (
        g_sync_edge => "positive"
      )
      port map
      (
        clk_i    => clk_20_i,
        rst_n_i  => '1',
        data_i   => trig_a(i),
        synced_o => trig_synced(i)
      );

    -- Deglitch synchronized trigger signal
    --
    -- Reset value is '1' to avoid pulses being counted by pulse counter on
    -- startup, when the board is in TTL-BAR repetition mode.
    cmp_inp_glitch_filt : gc_glitch_filt
      generic map
      (
        g_len => g_pgen_gf_len
      )
      port map
      (
        clk_i   => clk_20_i,
        rst_n_i => '1',
        dat_i   => trig_synced(i),
        dat_o   => trig_degl(i)
      );

    -- Now that we have a deglitched signal, generate the MUX to select between
    -- deglitched and direct channel input
    trig_chan(i) <= trig_a(i) when (gf_en_n_i = '1') else
                    trig_degl(i);

    -- The trigger to the pulse generator is multiplexed between manually triggered or
    -- channel input
    trig_pgen(i) <= trig_chan(i);

    -- Instantiate pulse generator block for the channel
    cmp_pulse_gen : conv_pulse_gen
      generic map
      (
        g_with_fixed_pwidth => g_pgen_fixed_width,
        g_pwidth            => g_pgen_pwidth,
        g_duty_cycle_div    => g_pgen_duty_cycle_div
      )
      port map
      (
        clk_i         => clk_20_i,
        rst_n_i       => rst_20_n,

        gf_en_n_i     => gf_en_n_i,

        en_i          => '1',

        trig_a_i      => trig_pgen(i),

        pulse_err_p_o => open,

        pulse_o       => pulse_outp(i)
      );

    -- Process to flash pulse LED when a pulse is output
    -- LED flash length: 26 ms
    p_pulse_led : process (clk_20_i) is
    begin
      if rising_edge(clk_20_i) then
        if (rst_20_n = '0') then
          pulse_outp_d0(i)      <= '0';
          pulse_outp_redge_p(i) <= '0';
          led_pulse_cnt(i)      <= (others => '0');
          led_pulse(i)          <= '0';
        else
          pulse_outp_d0(i)      <= pulse_outp(i);
          pulse_outp_redge_p(i) <= pulse_outp(i) and (not pulse_outp_d0(i));
          case led_pulse(i) is
            when '0' =>
              if (pulse_outp_redge_p(i) = '1') then
                led_pulse(i) <= '1';
              end if;
            when '1' =>
              led_pulse_cnt(i) <= led_pulse_cnt(i) + 1;
              if (led_pulse_cnt(i) = (led_pulse_cnt(i)'range => '1')) then
                led_pulse(i) <= '0';
              end if;
            when others =>
              led_pulse(i) <= '0';
          end case;
        end if;
      end if;
    end process p_pulse_led;

  end generate gen_pulse_chan_logic;

  -- Channel output assignments
  pulse_o     <= pulse_outp;
  led_pulse_o <= led_pulse;

  --============================================================================
  -- Bicolor LED matrix logic
  --============================================================================
  -- connect column, line & state signals outside
  cmp_bicolor_led_ctrl : gc_bicolor_led_ctrl
    generic map
    (
      g_NB_COLUMN    => g_bicolor_led_columns,
      g_NB_LINE      => g_bicolor_led_lines,
      g_clk_freq     => 20000000,
      g_refresh_rate => 250
    )
    port map
    (
      clk_i           => clk_20_i,
      rst_n_i         => rst_20_n,
      led_intensity_i => "1111111",
      led_state_i     => bicolor_led_state_i,
      column_o        => bicolor_led_col_o,
      line_o          => bicolor_led_line_o,
      line_oen_o      => bicolor_led_line_oen_o
    );

end architecture arch;
--==============================================================================
--  architecture end
--==============================================================================
